@@ -11,9 +11,10 @@ import struct
 import binascii
 
 from .exceptions import (
-    UnknownFrameError, InvalidPaddingError, InvalidFrameError
+    UnknownFrameError, InvalidPaddingError, InvalidFrameError, InvalidDataError
 )
 from .flags import Flag, Flags
+from typing import Optional, Tuple, List, Iterable, Any, Dict, Type
 
 
 # The maximum initial length of a frame. Some frames have shorter maximum
@@ -23,22 +24,36 @@ FRAME_MAX_LEN = (2 ** 14)
 # The maximum allowed length of a frame.
 FRAME_MAX_ALLOWED_LEN = (2 ** 24) - 1
 
+# Stream association enumerations.
+_STREAM_ASSOC_HAS_STREAM = "has-stream"
+_STREAM_ASSOC_NO_STREAM = "no-stream"
+_STREAM_ASSOC_EITHER = "either"
 
-class Frame(object):
+# Structs for packing and unpacking
+_STRUCT_HBBBL = struct.Struct(">HBBBL")
+_STRUCT_LL = struct.Struct(">LL")
+_STRUCT_HL = struct.Struct(">HL")
+_STRUCT_LB = struct.Struct(">LB")
+_STRUCT_L = struct.Struct(">L")
+_STRUCT_H = struct.Struct(">H")
+_STRUCT_B = struct.Struct(">B")
+
+
+class Frame:
     """
     The base class for all HTTP/2 frames.
     """
     #: The flags defined on this type of frame.
-    defined_flags = []
+    defined_flags: List[Flag] = []
 
     #: The byte used to define the type of the frame.
-    type = None
+    type: Optional[int] = None
 
     # If 'has-stream', the frame's stream_id must be non-zero. If 'no-stream',
     # it must be zero. If 'either', it's not checked.
-    stream_association = None
+    stream_association: Optional[str] = None
 
-    def __init__(self, stream_id, flags=()):
+    def __init__(self, stream_id: int, flags: Iterable[str] = ()) -> None:
         #: The stream identifier for the stream this frame was received on.
         #: Set to 0 for frames sent on the connection (stream-id 0).
         self.stream_id = stream_id
@@ -52,38 +67,76 @@ class Frame(object):
         for flag in flags:
             self.flags.add(flag)
 
-        if self.stream_association == 'has-stream' and not self.stream_id:
-            raise ValueError('Stream ID must be non-zero')
-        if self.stream_association == 'no-stream' and self.stream_id:
-            raise ValueError('Stream ID must be zero')
+        if (not self.stream_id and
+           self.stream_association == _STREAM_ASSOC_HAS_STREAM):
+            raise InvalidDataError(
+                'Stream ID must be non-zero for {}'.format(
+                    type(self).__name__,
+                )
+            )
+        if (self.stream_id and
+           self.stream_association == _STREAM_ASSOC_NO_STREAM):
+            raise InvalidDataError(
+                'Stream ID must be zero for {} with stream_id={}'.format(
+                    type(self).__name__,
+                    self.stream_id,
+                )
+            )
 
-    def __repr__(self):
-        flags = ", ".join(self.flags) or "None"
-        body = binascii.hexlify(self.serialize_body()).decode('ascii')
-        if len(body) > 20:
-            body = body[:20] + "..."
+    def __repr__(self) -> str:
         return (
-            "{type}(Stream: {stream}; Flags: {flags}): {body}"
+            "{}(stream_id={}, flags={}): {}"
         ).format(
-            type=type(self).__name__,
-            stream=self.stream_id,
-            flags=flags,
-            body=body
+            type(self).__name__,
+            self.stream_id,
+            repr(self.flags),
+            self._body_repr(),
         )
 
+    def _body_repr(self) -> str:
+        # More specific implementation may be provided by subclasses of Frame.
+        # This fallback shows the serialized (and truncated) body content.
+        return _raw_data_repr(self.serialize_body())
+
     @staticmethod
-    def parse_frame_header(header):
+    def explain(data: memoryview) -> Tuple["Frame", int]:
+        """
+        Takes a bytestring and tries to parse a single frame and print it.
+
+        This function is only provided for debugging purposes.
+
+        :param data: A memoryview object containing the raw data of at least
+                     one complete frame (header and body).
+
+        .. versionadded:: 6.0.0
+        """
+        frame, length = Frame.parse_frame_header(data[:9])
+        frame.parse_body(data[9:9 + length])
+        print(frame)
+        return frame, length
+
+    @staticmethod
+    def parse_frame_header(header: memoryview, strict: bool = False) -> Tuple["Frame", int]:
         """
         Takes a 9-byte frame header and returns a tuple of the appropriate
         Frame object and the length that needs to be read from the socket.
 
         This populates the flags field, and determines how long the body is.
 
+        :param header: A memoryview object containing the 9-byte frame header
+                       data of a frame. Must not contain more or less.
+
+        :param strict: Whether to raise an exception when encountering a frame
+            not defined by spec and implemented by hyperframe.
+
         :raises hyperframe.exceptions.UnknownFrameError: If a frame of unknown
             type is received.
+
+        .. versionchanged:: 5.0.0
+            Added :param:`strict` to accommodate :class:`ExtensionFrame`
         """
         try:
-            fields = struct.unpack("!HBBBL", header)
+            fields = _STRUCT_HBBBL.unpack(header)
         except struct.error:
             raise InvalidFrameError("Invalid frame header")
 
@@ -91,23 +144,26 @@ class Frame(object):
         length = (fields[0] << 8) + fields[1]
         type = fields[2]
         flags = fields[3]
-        stream_id = fields[4]
+        stream_id = fields[4] & 0x7FFFFFFF
 
-        if type not in FRAMES:
-            raise UnknownFrameError(type, length)
+        try:
+            frame = FRAMES[type](stream_id)
+        except KeyError:
+            if strict:
+                raise UnknownFrameError(type, length)
+            frame = ExtensionFrame(type=type, stream_id=stream_id)
 
-        frame = FRAMES[type](stream_id)
         frame.parse_flags(flags)
         return (frame, length)
 
-    def parse_flags(self, flag_byte):
+    def parse_flags(self, flag_byte: int) -> Flags:
         for flag, flag_bit in self.defined_flags:
             if flag_byte & flag_bit:
                 self.flags.add(flag)
 
         return self.flags
 
-    def serialize(self):
+    def serialize(self) -> bytes:
         """
         Convert a frame into a bytestring, representing the serialized form of
         the frame.
@@ -123,10 +179,9 @@ class Frame(object):
             if flag in self.flags:
                 flags |= flag_bit
 
-        header = struct.pack(
-            "!HBBBL",
-            (self.body_len & 0xFFFF00) >> 8,  # Length spread over top 24 bits
-            self.body_len & 0x0000FF,
+        header = _STRUCT_HBBBL.pack(
+            (self.body_len >> 8) & 0xFFFF,  # Length spread over top 24 bits
+            self.body_len & 0xFF,
             self.type,
             flags,
             self.stream_id & 0x7FFFFFFF  # Stream ID is 32 bits.
@@ -134,10 +189,10 @@ class Frame(object):
 
         return header + body
 
-    def serialize_body(self):
+    def serialize_body(self) -> bytes:
         raise NotImplementedError()
 
-    def parse_body(self, data):
+    def parse_body(self, data: memoryview) -> None:
         """
         Given the body of a frame, parses it into frame data. This populates
         the non-header parts of the frame: that is, it does not populate the
@@ -151,24 +206,24 @@ class Frame(object):
         raise NotImplementedError()
 
 
-class Padding(object):
+class Padding:
     """
     Mixin for frames that contain padding. Defines extra fields that can be
     used and set by frames that can be padded.
     """
-    def __init__(self, stream_id, pad_length=0, **kwargs):
-        super(Padding, self).__init__(stream_id, **kwargs)
+    def __init__(self, stream_id: int, pad_length: int = 0, **kwargs: Any) -> None:
+        super().__init__(stream_id, **kwargs)  # type: ignore
 
         #: The length of the padding to use.
         self.pad_length = pad_length
 
-    def serialize_padding_data(self):
-        if 'PADDED' in self.flags:
-            return struct.pack('!B', self.pad_length)
+    def serialize_padding_data(self) -> bytes:
+        if 'PADDED' in self.flags:  # type: ignore
+            return _STRUCT_B.pack(self.pad_length)
         return b''
 
-    def parse_padding_data(self, data):
-        if 'PADDED' in self.flags:
+    def parse_padding_data(self, data: memoryview) -> int:
+        if 'PADDED' in self.flags:  # type: ignore
             try:
                 self.pad_length = struct.unpack('!B', data[:1])[0]
             except struct.error:
@@ -176,23 +231,30 @@ class Padding(object):
             return 1
         return 0
 
+    #: .. deprecated:: 5.2.1
+    #:    Use self.pad_length instead.
     @property
-    def total_padding(self):
+    def total_padding(self) -> int:  # pragma: no cover
+        import warnings
+        warnings.warn(
+            "total_padding contains the same information as pad_length.",
+            DeprecationWarning
+        )
         return self.pad_length
 
 
-class Priority(object):
+class Priority:
     """
     Mixin for frames that contain priority data. Defines extra fields that can
     be used and set by frames that contain priority data.
     """
     def __init__(self,
-                 stream_id,
-                 depends_on=0x0,
-                 stream_weight=0x0,
-                 exclusive=False,
-                 **kwargs):
-        super(Priority, self).__init__(stream_id, **kwargs)
+                 stream_id: int,
+                 depends_on: int = 0x0,
+                 stream_weight: int = 0x0,
+                 exclusive: bool = False,
+                 **kwargs: Any) -> None:
+        super().__init__(stream_id, **kwargs)  # type: ignore
 
         #: The stream ID of the stream on which this stream depends.
         self.depends_on = depends_on
@@ -203,25 +265,20 @@ class Priority(object):
         #: Whether the exclusive bit was set.
         self.exclusive = exclusive
 
-    def serialize_priority_data(self):
-        return struct.pack(
-            "!LB",
-            self.depends_on | (int(self.exclusive) << 31),
+    def serialize_priority_data(self) -> bytes:
+        return _STRUCT_LB.pack(
+            self.depends_on + (0x80000000 if self.exclusive else 0),
             self.stream_weight
         )
 
-    def parse_priority_data(self, data):
-        MASK = 0x80000000
-
+    def parse_priority_data(self, data: memoryview) -> int:
         try:
-            self.depends_on, self.stream_weight = struct.unpack(
-                "!LB", data[:5]
-            )
+            self.depends_on, self.stream_weight = _STRUCT_LB.unpack(data[:5])
         except struct.error:
             raise InvalidFrameError("Invalid Priority data")
 
-        self.exclusive = bool(self.depends_on & MASK)
-        self.depends_on &= ~MASK
+        self.exclusive = True if self.depends_on >> 31 else False
+        self.depends_on &= 0x7FFFFFFF
         return 5
 
 
@@ -240,36 +297,42 @@ class DataFrame(Padding, Frame):
     #: The type byte for data frames.
     type = 0x0
 
-    stream_association = 'has-stream'
+    stream_association = _STREAM_ASSOC_HAS_STREAM
 
-    def __init__(self, stream_id, data=b'', **kwargs):
-        super(DataFrame, self).__init__(stream_id, **kwargs)
+    def __init__(self, stream_id: int, data: bytes = b'', **kwargs: Any) -> None:
+        super().__init__(stream_id, **kwargs)
 
         #: The data contained on this frame.
         self.data = data
 
-    def serialize_body(self):
+    def serialize_body(self) -> bytes:
         padding_data = self.serialize_padding_data()
-        padding = b'\0' * self.total_padding
+        padding = b'\0' * self.pad_length
+        if isinstance(self.data, memoryview):
+            self.data = self.data.tobytes()
         return b''.join([padding_data, self.data, padding])
 
-    def parse_body(self, data):
+    def parse_body(self, data: memoryview) -> None:
         padding_data_length = self.parse_padding_data(data)
         self.data = (
-            data[padding_data_length:len(data)-self.total_padding].tobytes()
+            data[padding_data_length:len(data)-self.pad_length].tobytes()
         )
         self.body_len = len(data)
 
-        if self.total_padding and self.total_padding >= self.body_len:
+        if self.pad_length and self.pad_length >= self.body_len:
             raise InvalidPaddingError("Padding is too long.")
 
     @property
-    def flow_controlled_length(self):
+    def flow_controlled_length(self) -> int:
         """
         The length of the frame that needs to be accounted for when considering
         flow control.
         """
-        padding_len = self.total_padding + 1 if self.total_padding else 0
+        padding_len = 0
+        if 'PADDED' in self.flags:
+            # Account for extra 1-byte padding length field, which is still
+            # present if possibly zero-valued.
+            padding_len = self.pad_length + 1
         return len(self.data) + padding_len
 
 
@@ -280,19 +343,32 @@ class PriorityFrame(Priority, Frame):
     reprioritisation of existing streams.
     """
     #: The flags defined for PRIORITY frames.
-    defined_flags = []
+    defined_flags: List[Flag] = []
 
     #: The type byte defined for PRIORITY frames.
     type = 0x02
 
-    stream_association = 'has-stream'
+    stream_association = _STREAM_ASSOC_HAS_STREAM
 
-    def serialize_body(self):
+    def _body_repr(self) -> str:
+        return "exclusive={}, depends_on={}, stream_weight={}".format(
+            self.exclusive,
+            self.depends_on,
+            self.stream_weight
+        )
+
+    def serialize_body(self) -> bytes:
         return self.serialize_priority_data()
 
-    def parse_body(self, data):
+    def parse_body(self, data: memoryview) -> None:
+        if len(data) > 5:
+            raise InvalidFrameError(
+                "PRIORITY must have 5 byte body: actual length %s." %
+                len(data)
+            )
+
         self.parse_priority_data(data)
-        self.body_len = len(data)
+        self.body_len = 5
 
 
 class RstStreamFrame(Frame):
@@ -305,23 +381,28 @@ class RstStreamFrame(Frame):
     occurred.
     """
     #: The flags defined for RST_STREAM frames.
-    defined_flags = []
+    defined_flags: List[Flag] = []
 
     #: The type byte defined for RST_STREAM frames.
     type = 0x03
 
-    stream_association = 'has-stream'
+    stream_association = _STREAM_ASSOC_HAS_STREAM
 
-    def __init__(self, stream_id, error_code=0, **kwargs):
-        super(RstStreamFrame, self).__init__(stream_id, **kwargs)
+    def __init__(self, stream_id: int, error_code: int = 0, **kwargs: Any) -> None:
+        super().__init__(stream_id, **kwargs)
 
         #: The error code used when resetting the stream.
         self.error_code = error_code
 
-    def serialize_body(self):
-        return struct.pack("!L", self.error_code)
+    def _body_repr(self) -> str:
+        return "error_code={}".format(
+            self.error_code,
+        )
 
-    def parse_body(self, data):
+    def serialize_body(self) -> bytes:
+        return _STRUCT_L.pack(self.error_code)
+
+    def parse_body(self, data: memoryview) -> None:
         if len(data) != 4:
             raise InvalidFrameError(
                 "RST_STREAM must have 4 byte body: actual length %s." %
@@ -329,11 +410,11 @@ class RstStreamFrame(Frame):
             )
 
         try:
-            self.error_code = struct.unpack("!L", data)[0]
+            self.error_code = _STRUCT_L.unpack(data)[0]
         except struct.error:  # pragma: no cover
             raise InvalidFrameError("Invalid RST_STREAM body")
 
-        self.body_len = len(data)
+        self.body_len = 4
 
 
 class SettingsFrame(Frame):
@@ -354,7 +435,7 @@ class SettingsFrame(Frame):
     #: The type byte defined for SETTINGS frames.
     type = 0x04
 
-    stream_association = 'no-stream'
+    stream_association = _STREAM_ASSOC_NO_STREAM
 
     # We need to define the known settings, they may as well be class
     # attributes.
@@ -370,31 +451,47 @@ class SettingsFrame(Frame):
     MAX_FRAME_SIZE = 0x05
     #: The byte that signals the SETTINGS_MAX_HEADER_LIST_SIZE setting.
     MAX_HEADER_LIST_SIZE = 0x06
+    #: The byte that signals SETTINGS_ENABLE_CONNECT_PROTOCOL setting.
+    ENABLE_CONNECT_PROTOCOL = 0x08
 
-    def __init__(self, stream_id=0, settings=None, **kwargs):
-        super(SettingsFrame, self).__init__(stream_id, **kwargs)
+    def __init__(self, stream_id: int = 0, settings: Optional[Dict[int, int]] = None, **kwargs: Any) -> None:
+        super().__init__(stream_id, **kwargs)
 
         if settings and "ACK" in kwargs.get("flags", ()):
-            raise ValueError("Settings must be empty if ACK flag is set.")
+            raise InvalidDataError(
+                "Settings must be empty if ACK flag is set."
+            )
 
         #: A dictionary of the setting type byte to the value of the setting.
         self.settings = settings or {}
 
-    def serialize_body(self):
-        settings = [struct.pack("!HL", setting & 0xFF, value)
-                    for setting, value in self.settings.items()]
-        return b''.join(settings)
+    def _body_repr(self) -> str:
+        return "settings={}".format(
+            self.settings,
+        )
 
-    def parse_body(self, data):
+    def serialize_body(self) -> bytes:
+        return b''.join([_STRUCT_HL.pack(setting & 0xFF, value)
+                         for setting, value in self.settings.items()])
+
+    def parse_body(self, data: memoryview) -> None:
+        if 'ACK' in self.flags and len(data) > 0:
+            raise InvalidDataError(
+                "SETTINGS ack frame must not have payload: got %s bytes" %
+                len(data)
+            )
+
+        body_len = 0
         for i in range(0, len(data), 6):
             try:
-                name, value = struct.unpack("!HL", data[i:i+6])
+                name, value = _STRUCT_HL.unpack(data[i:i+6])
             except struct.error:
                 raise InvalidFrameError("Invalid SETTINGS body")
 
             self.settings[name] = value
+            body_len += 6
 
-        self.body_len = len(data)
+        self.body_len = body_len
 
 
 class PushPromiseFrame(Padding, Frame):
@@ -411,10 +508,10 @@ class PushPromiseFrame(Padding, Frame):
     #: The type byte defined for PUSH_PROMISE frames.
     type = 0x05
 
-    stream_association = 'has-stream'
+    stream_association = _STREAM_ASSOC_HAS_STREAM
 
-    def __init__(self, stream_id, promised_stream_id=0, data=b'', **kwargs):
-        super(PushPromiseFrame, self).__init__(stream_id, **kwargs)
+    def __init__(self, stream_id: int, promised_stream_id: int = 0, data: bytes = b'', **kwargs: Any) -> None:
+        super().__init__(stream_id, **kwargs)
 
         #: The stream ID that is promised by this frame.
         self.promised_stream_id = promised_stream_id
@@ -423,26 +520,40 @@ class PushPromiseFrame(Padding, Frame):
         #: stream.
         self.data = data
 
-    def serialize_body(self):
+    def _body_repr(self) -> str:
+        return "promised_stream_id={}, data={}".format(
+            self.promised_stream_id,
+            _raw_data_repr(self.data),
+        )
+
+    def serialize_body(self) -> bytes:
         padding_data = self.serialize_padding_data()
-        padding = b'\0' * self.total_padding
-        data = struct.pack("!L", self.promised_stream_id)
+        padding = b'\0' * self.pad_length
+        data = _STRUCT_L.pack(self.promised_stream_id)
         return b''.join([padding_data, data, self.data, padding])
 
-    def parse_body(self, data):
+    def parse_body(self, data: memoryview) -> None:
         padding_data_length = self.parse_padding_data(data)
 
         try:
-            self.promised_stream_id = struct.unpack(
-                "!L", data[padding_data_length:padding_data_length + 4]
+            self.promised_stream_id = _STRUCT_L.unpack(
+                data[padding_data_length:padding_data_length + 4]
             )[0]
         except struct.error:
             raise InvalidFrameError("Invalid PUSH_PROMISE body")
 
-        self.data = data[padding_data_length + 4:].tobytes()
+        self.data = (
+            data[padding_data_length + 4:len(data)-self.pad_length].tobytes()
+        )
         self.body_len = len(data)
 
-        if self.total_padding and self.total_padding >= self.body_len:
+        if self.promised_stream_id == 0 or self.promised_stream_id % 2 != 0:
+            raise InvalidDataError(
+                "Invalid PUSH_PROMISE promised stream id: %s" %
+                self.promised_stream_id
+            )
+
+        if self.pad_length and self.pad_length >= self.body_len:
             raise InvalidPaddingError("Padding is too long.")
 
 
@@ -458,18 +569,23 @@ class PingFrame(Frame):
     #: The type byte defined for PING frames.
     type = 0x06
 
-    stream_association = 'no-stream'
+    stream_association = _STREAM_ASSOC_NO_STREAM
 
-    def __init__(self, stream_id=0, opaque_data=b'', **kwargs):
-        super(PingFrame, self).__init__(stream_id, **kwargs)
+    def __init__(self, stream_id: int = 0, opaque_data: bytes = b'', **kwargs: Any) -> None:
+        super().__init__(stream_id, **kwargs)
 
         #: The opaque data sent in this PING frame, as a bytestring.
         self.opaque_data = opaque_data
 
-    def serialize_body(self):
+    def _body_repr(self) -> str:
+        return "opaque_data={!r}".format(
+            self.opaque_data,
+        )
+
+    def serialize_body(self) -> bytes:
         if len(self.opaque_data) > 8:
             raise InvalidFrameError(
-                "PING frame may not have more than 8 bytes of data, got %s" %
+                "PING frame may not have more than 8 bytes of data, got %r" %
                 self.opaque_data
             )
 
@@ -477,14 +593,14 @@ class PingFrame(Frame):
         data += b'\x00' * (8 - len(self.opaque_data))
         return data
 
-    def parse_body(self, data):
+    def parse_body(self, data: memoryview) -> None:
         if len(data) != 8:
             raise InvalidFrameError(
                 "PING frame must have 8 byte length: got %s" % len(data)
             )
 
         self.opaque_data = data.tobytes()
-        self.body_len = len(data)
+        self.body_len = 8
 
 
 class GoAwayFrame(Frame):
@@ -495,20 +611,20 @@ class GoAwayFrame(Frame):
     connection.
     """
     #: The flags defined for GOAWAY frames.
-    defined_flags = []
+    defined_flags: List[Flag] = []
 
     #: The type byte defined for GOAWAY frames.
     type = 0x07
 
-    stream_association = 'no-stream'
+    stream_association = _STREAM_ASSOC_NO_STREAM
 
     def __init__(self,
-                 stream_id=0,
-                 last_stream_id=0,
-                 error_code=0,
-                 additional_data=b'',
-                 **kwargs):
-        super(GoAwayFrame, self).__init__(stream_id, **kwargs)
+                 stream_id: int = 0,
+                 last_stream_id: int = 0,
+                 error_code: int = 0,
+                 additional_data: bytes = b'',
+                 **kwargs: Any) -> None:
+        super().__init__(stream_id, **kwargs)
 
         #: The last stream ID definitely seen by the remote peer.
         self.last_stream_id = last_stream_id
@@ -519,9 +635,15 @@ class GoAwayFrame(Frame):
         #: Any additional data sent in the GOAWAY.
         self.additional_data = additional_data
 
-    def serialize_body(self):
-        data = struct.pack(
-            "!LL",
+    def _body_repr(self) -> str:
+        return "last_stream_id={}, error_code={}, additional_data={!r}".format(
+            self.last_stream_id,
+            self.error_code,
+            self.additional_data,
+        )
+
+    def serialize_body(self) -> bytes:
+        data = _STRUCT_LL.pack(
             self.last_stream_id & 0x7FFFFFFF,
             self.error_code
         )
@@ -529,10 +651,10 @@ class GoAwayFrame(Frame):
 
         return data
 
-    def parse_body(self, data):
+    def parse_body(self, data: memoryview) -> None:
         try:
-            self.last_stream_id, self.error_code = struct.unpack(
-                "!LL", data[:8]
+            self.last_stream_id, self.error_code = _STRUCT_LL.unpack(
+                data[:8]
             )
         except struct.error:
             raise InvalidFrameError("Invalid GOAWAY body.")
@@ -557,29 +679,45 @@ class WindowUpdateFrame(Frame):
     original sender.
     """
     #: The flags defined for WINDOW_UPDATE frames.
-    defined_flags = []
+    defined_flags: List[Flag] = []
 
     #: The type byte defined for WINDOW_UPDATE frames.
     type = 0x08
 
-    stream_association = 'either'
+    stream_association = _STREAM_ASSOC_EITHER
 
-    def __init__(self, stream_id, window_increment=0, **kwargs):
-        super(WindowUpdateFrame, self).__init__(stream_id, **kwargs)
+    def __init__(self, stream_id: int, window_increment: int = 0, **kwargs: Any) -> None:
+        super().__init__(stream_id, **kwargs)
 
         #: The amount the flow control window is to be incremented.
         self.window_increment = window_increment
 
-    def serialize_body(self):
-        return struct.pack("!L", self.window_increment & 0x7FFFFFFF)
+    def _body_repr(self) -> str:
+        return "window_increment={}".format(
+            self.window_increment,
+        )
+
+    def serialize_body(self) -> bytes:
+        return _STRUCT_L.pack(self.window_increment & 0x7FFFFFFF)
+
+    def parse_body(self, data: memoryview) -> None:
+        if len(data) > 4:
+            raise InvalidFrameError(
+                "WINDOW_UPDATE frame must have 4 byte length: got %s" %
+                len(data)
+            )
 
-    def parse_body(self, data):
         try:
-            self.window_increment = struct.unpack("!L", data)[0]
+            self.window_increment = _STRUCT_L.unpack(data)[0]
         except struct.error:
             raise InvalidFrameError("Invalid WINDOW_UPDATE body")
 
-        self.body_len = len(data)
+        if not 1 <= self.window_increment <= 2**31-1:
+            raise InvalidDataError(
+                "WINDOW_UPDATE increment must be between 1 to 2^31-1"
+            )
+
+        self.body_len = 4
 
 
 class HeadersFrame(Padding, Priority, Frame):
@@ -605,17 +743,25 @@ class HeadersFrame(Padding, Priority, Frame):
     #: The type byte defined for HEADERS frames.
     type = 0x01
 
-    stream_association = 'has-stream'
+    stream_association = _STREAM_ASSOC_HAS_STREAM
 
-    def __init__(self, stream_id, data=b'', **kwargs):
-        super(HeadersFrame, self).__init__(stream_id, **kwargs)
+    def __init__(self, stream_id: int, data: bytes = b'', **kwargs: Any) -> None:
+        super().__init__(stream_id, **kwargs)
 
         #: The HPACK-encoded header block.
         self.data = data
 
-    def serialize_body(self):
+    def _body_repr(self) -> str:
+        return "exclusive={}, depends_on={}, stream_weight={}, data={}".format(
+            self.exclusive,
+            self.depends_on,
+            self.stream_weight,
+            _raw_data_repr(self.data),
+        )
+
+    def serialize_body(self) -> bytes:
         padding_data = self.serialize_padding_data()
-        padding = b'\0' * self.total_padding
+        padding = b'\0' * self.pad_length
 
         if 'PRIORITY' in self.flags:
             priority_data = self.serialize_priority_data()
@@ -624,7 +770,7 @@ class HeadersFrame(Padding, Priority, Frame):
 
         return b''.join([padding_data, priority_data, self.data, padding])
 
-    def parse_body(self, data):
+    def parse_body(self, data: memoryview) -> None:
         padding_data_length = self.parse_padding_data(data)
         data = data[padding_data_length:]
 
@@ -635,10 +781,10 @@ class HeadersFrame(Padding, Priority, Frame):
 
         self.body_len = len(data)
         self.data = (
-            data[priority_data_length:len(data)-self.total_padding].tobytes()
+            data[priority_data_length:len(data)-self.pad_length].tobytes()
         )
 
-        if self.total_padding and self.total_padding >= self.body_len:
+        if self.pad_length and self.pad_length >= self.body_len:
             raise InvalidPaddingError("Padding is too long.")
 
 
@@ -658,18 +804,23 @@ class ContinuationFrame(Frame):
     #: The type byte defined for CONTINUATION frames.
     type = 0x09
 
-    stream_association = 'has-stream'
+    stream_association = _STREAM_ASSOC_HAS_STREAM
 
-    def __init__(self, stream_id, data=b'', **kwargs):
-        super(ContinuationFrame, self).__init__(stream_id, **kwargs)
+    def __init__(self, stream_id: int, data: bytes = b'', **kwargs: Any) -> None:
+        super().__init__(stream_id, **kwargs)
 
         #: The HPACK-encoded header block.
         self.data = data
 
-    def serialize_body(self):
+    def _body_repr(self) -> str:
+        return "data={}".format(
+            _raw_data_repr(self.data),
+        )
+
+    def serialize_body(self) -> bytes:
         return self.data
 
-    def parse_body(self, data):
+    def parse_body(self, data: memoryview) -> None:
         self.data = data.tobytes()
         self.body_len = len(data)
 
@@ -691,25 +842,31 @@ class AltSvcFrame(Frame):
     """
     type = 0xA
 
-    stream_association = 'both'
+    stream_association = _STREAM_ASSOC_EITHER
 
-    def __init__(self, stream_id, origin=b'', field=b'', **kwargs):
-        super(AltSvcFrame, self).__init__(stream_id, **kwargs)
+    def __init__(self, stream_id: int, origin: bytes = b'', field: bytes = b'', **kwargs: Any) -> None:
+        super().__init__(stream_id, **kwargs)
 
         if not isinstance(origin, bytes):
-            raise ValueError("AltSvc origin must be bytestring.")
+            raise InvalidDataError("AltSvc origin must be bytestring.")
         if not isinstance(field, bytes):
-            raise ValueError("AltSvc field must be a bytestring.")
+            raise InvalidDataError("AltSvc field must be a bytestring.")
         self.origin = origin
         self.field = field
 
-    def serialize_body(self):
-        origin_len = struct.pack("!H", len(self.origin))
+    def _body_repr(self) -> str:
+        return "origin={!r}, field={!r}".format(
+            self.origin,
+            self.field,
+        )
+
+    def serialize_body(self) -> bytes:
+        origin_len = _STRUCT_H.pack(len(self.origin))
         return b''.join([origin_len, self.origin, self.field])
 
-    def parse_body(self, data):
+    def parse_body(self, data: memoryview) -> None:
         try:
-            origin_len = struct.unpack("!H", data[0:2])[0]
+            origin_len = _STRUCT_H.unpack(data[0:2])[0]
             self.origin = data[2:2+origin_len].tobytes()
 
             if len(self.origin) != origin_len:
@@ -722,7 +879,79 @@ class AltSvcFrame(Frame):
         self.body_len = len(data)
 
 
-_FRAME_CLASSES = [
+class ExtensionFrame(Frame):
+    """
+    ExtensionFrame is used to wrap frames which are not natively interpretable
+    by hyperframe.
+
+    Although certain byte prefixes are ordained by specification to have
+    certain contextual meanings, frames with other prefixes are not prohibited,
+    and may be used to communicate arbitrary meaning between HTTP/2 peers.
+
+    Thus, hyperframe, rather than raising an exception when such a frame is
+    encountered, wraps it in a generic frame to be properly acted upon by
+    upstream consumers which might have additional context on how to use it.
+
+    .. versionadded:: 5.0.0
+    """
+
+    stream_association = _STREAM_ASSOC_EITHER
+
+    def __init__(self, type: int, stream_id: int, flag_byte: int = 0x0, body: bytes = b'', **kwargs: Any) -> None:
+        super().__init__(stream_id, **kwargs)
+        self.type = type
+        self.flag_byte = flag_byte
+        self.body = body
+
+    def _body_repr(self) -> str:
+        return "type={}, flag_byte={}, body={}".format(
+            self.type,
+            self.flag_byte,
+            _raw_data_repr(self.body),
+        )
+
+    def parse_flags(self, flag_byte: int) -> None:  # type: ignore
+        """
+        For extension frames, we parse the flags by just storing a flag byte.
+        """
+        self.flag_byte = flag_byte
+
+    def parse_body(self, data: memoryview) -> None:
+        self.body = data.tobytes()
+        self.body_len = len(data)
+
+    def serialize(self) -> bytes:
+        """
+        A broad override of the serialize method that ensures that the data
+        comes back out exactly as it came in. This should not be used in most
+        user code: it exists only as a helper method if frames need to be
+        reconstituted.
+        """
+        # Build the frame header.
+        # First, get the flags.
+        flags = self.flag_byte
+
+        header = _STRUCT_HBBBL.pack(
+            (self.body_len >> 8) & 0xFFFF,  # Length spread over top 24 bits
+            self.body_len & 0xFF,
+            self.type,
+            flags,
+            self.stream_id & 0x7FFFFFFF  # Stream ID is 32 bits.
+        )
+
+        return header + self.body
+
+
+def _raw_data_repr(data: Optional[bytes]) -> str:
+    if not data:
+        return "None"
+    r = binascii.hexlify(data).decode('ascii')
+    if len(r) > 20:
+        r = r[:20] + "..."
+    return "<hex:" + r + ">"
+
+
+_FRAME_CLASSES: List[Type[Frame]] = [
     DataFrame,
     HeadersFrame,
     PriorityFrame,
